#include <J3ML/Algorithm/GJK.hpp>
#include <J3ML/Geometry/Capsule.hpp>
#include <J3ML/Geometry/AABB.hpp>
#include <J3ML/Geometry/Sphere.hpp>
#include <J3ML/Geometry/Polygon.hpp>

namespace J3ML::Geometry
{

    Capsule::Capsule() : l() {}

    Capsule::Capsule(const LineSegment &endPoints, float radius)
    :l(endPoints), r(radius)
    {
    }

    Capsule::Capsule(const Vector3 &bottomPoint, const Vector3 &topPoint, float radius)
    :l(bottomPoint, topPoint), r(radius)
    {
    }


    AABB Capsule::MinimalEnclosingAABB() const
    {
        Vector3 d = Vector3(r, r, r);
        AABB aabb(Vector3::Min(l.A, l.B) - d, Vector3::Max(l.A, l.B) + d);
        return aabb;
    }

    void Capsule::ProjectToAxis(const Vector3 &direction, float &outMin, float &outMax) const {
        outMin = Vector3::Dot(direction, l.A);
        outMax = Vector3::Dot(direction, l.B);
        if (outMax < outMin)
            Swap(outMin, outMax);

        // The following requires that direction is normalized, otherwise we would have to sub/add 'r * direction.Length()', but
        // don't want to do that for performance reasons.
        assert(direction.IsNormalized());
        outMin -= r;
        outMax += r;
    }

    bool Capsule::Intersects(const Ray &ray) const
    {
        return l.Distance(ray) <= r;
    }

    bool Capsule::Intersects(const Line &line) const
    {
        return l.Distance(line) <= r;
    }

    bool Capsule::Intersects(const LineSegment &lineSegment) const
    {
        return l.Distance(lineSegment) <= r;
    }

    bool Capsule::Intersects(const Plane &plane) const {
        return l.Distance(plane) <= r;
    }

    bool Capsule::Intersects(const AABB &aabb) const
    {
        return Algorithms::FloatingPointOffsetedGJKIntersect(*this, aabb);
        return false;
    }

    bool Capsule::Intersects(const OBB &obb) const
    {
        return Algorithms::GJKIntersect(*this, obb);
    }

/// [groupSyntax]
    bool Capsule::Intersects(const Sphere &sphere) const
    {
        float R = r + sphere.Radius;
        return l.DistanceSq(sphere.Position) <= R*R;
    }

/// [groupSyntax]
    bool Capsule::Intersects(const Capsule &capsule) const
    {
        float R = r + capsule.r;
        return l.DistanceSq(capsule.l) <= R*R;
    }

    bool Capsule::Intersects(const Triangle &triangle) const
    {
        Vector3 thisPoint;
        Vector3 trianglePoint = triangle.ClosestPoint(l, &thisPoint);
        return thisPoint.DistanceSq(trianglePoint) <= r*r;
    }

    bool Capsule::Intersects(const Polygon &polygon) const
    {
        return polygon.Intersects(*this);
    }

    bool Capsule::Intersects(const Frustum &frustum) const
    {
        return frustum.Intersects(*this);
    }

    bool Capsule::Intersects(const Polyhedron &polyhedron) const
    {
        return polyhedron.Intersects(*this);
    }

    Sphere Capsule::SphereA() const {
        return Sphere(l.A, r);
    }

    Sphere Capsule::SphereB() const {
        return Sphere(l.B, r);
    }

    Vector3 Capsule::ExtremePoint(const Vector3 &direction) const {
        float len = direction.Length();
        assert(len > 0.f);
        return (Vector3::Dot(direction, l.B - l.A) >= 0.f ? l.B : l.A) + direction * (r / len);
    }

    Vector3 Capsule::ExtremePoint(const Vector3 &direction, float &projectionDistance) const {
        Vector3 extremePoint = ExtremePoint(direction);
        projectionDistance = extremePoint.Dot(direction);
        return extremePoint;
    }

    bool Capsule::IsDegenerate() const {
        return r <= 0.f; // Ask Bill :p
    }

    float Capsule::Height() const {
        return LineLength() + Diameter();
    }

    Capsule Capsule::Translated(const Vector3 &offset) const
    {
        return Capsule(l.A + offset, l.B + offset, r);
    }

    Vector3 Capsule::Bottom() const { return l.A - UpDirection() * r;}

    Vector3 Capsule::Center() const { return l.CenterPoint();}

    Vector3 Capsule::Centroid() const { return Center(); }

    float Capsule::Diameter() const { return 2.f * r;}

    float Capsule::LineLength() const { return l.Length(); }

    void Capsule::SetFrom(const Sphere &s) {
        l = LineSegment(s.Position, s.Position);
        r = s.Radius;
    }


    void Capsule::SetDegenerate() {
        r = -1.f;
    }

    Vector3 Capsule::AnyPointFast() const { return l.A; }


    Vector3 Capsule::UniformPointPerhapsInside(float height, float x, float y) const {
        return MinimalEnclosingOBB().PointInside(height, x, y);
    }


    Vector3 Capsule::UpDirection() const
    {
        Vector3 d = l.B - l.A;
        d.Normalize(); // Will always result in a normalized vector, even if l.a == l.b.
        return d;
    }

    Vector3 Capsule::Top() const
    {
        return l.B + UpDirection() * r;
    }

    float Capsule::Volume() const
    {
        return Math::Pi * r * r * LineLength() + 4.f * Math::Pi * r * r * r / 3.f;
    }

    float Capsule::SurfaceArea() const
    {
        return 2.f * Math::Pi * r * LineLength() + 4.f * Math::Pi * r * r;
    }

    OBB Capsule::MinimalEnclosingOBB() const {
        OBB obb;
        obb.axis[0] = UpDirection();
        obb.axis[0].PerpendicularBasis(obb.axis[1], obb.axis[2]);
        obb.pos = Center();
        obb.r[0] = Height() * 0.5f;
        obb.r[1] = r;
        obb.r[2] = r;
        return obb;
    }


}